*********************************************************************
*procedure ERRLOG
*      (c) B. Peisch for Peisch Custom Software, Inc., 1990-2002

**** FAT CLIENT VERSION OF ERROR LOGGING ****
* Compatible with all versions of Visual FoxPro

parameters tnDataSess, tnLineNo, tcUserID, tcErr_Mess

* COPYRIGHT:
*   The original author maintains the copyright for this utility under the
*   laws governing intellectual property in the United States.
*
*   You may use, modify, and distribute this utility, including source code,
*   as long as the original copyright information remains intact.  You may
*   not sell or license this utility unless included in an application.
*   Any advertisment of this Utility as a feature must also credit the
*   original author.

****************************************************************
* Include this utility as a procedure in your applications or by uncommenting
* the *procedure ERRLOG line.
* This program will run as a standalone program as is.
*
* Purpose of Errlog:
*   The most common use of Errlog is to trap any situation where the user
*   would normally see a FoxPro error message with the "Cancel, Ignore, or
*   Suspend" options.  In this situation, Errlog will log system information
*   at the time of the error and quit.  You may use the information recorded
*   by Errlog to assist you in debugging.
*
*   Additionally, you may use Errlog to record events in your application.
*   Using the optional Err_Mess parameter tells Errlog you are recording an
*   event.  When an event is recorded, no information is shown to the user.
*   System information is recorded and returns to running your application.
*
*   The file ERRORLOG.DBF stores all the errors that have been logged.  The
*   file will be created if it does not exist.  To clear the error log, erase
*   the files ERRORLOG.DBF and ERRORLOG.FPT.
*
* Calling Errlog:
*   System Errors:
*   Parameters are optional, but it is strongly recommended that you pass
*   at least the line number using the built in FoxPro LineNo() function.
*   Place the following command in your program to turn on error logging.
*   On error do ERRLOG with ;
*			iif(type('_screen.activeform') <> 'O', 0, _screen.activeform.DataSessionId), ;
*			lineno()
*
*   If you have a mechanism for capturing the user, you can also pass a user ID of up to
*   20 characters.  Use Sys(0) or a memory variable to pass a user ID.
*       Example:   On Error do ERRLOG with 0, Lineno(), Sys(0)
*            or:   On Error do ERRLOG with 0, Lineno(), UserID
*         where UserID is a variable set up by your application.
*
*   Most of the time you won't want to call error logging with a message.
*   Leaving the tcErr_Mess parameter empty indicates a system error and Errlog
*   will quit after recording information.
*
*   Recording Events:
*   You can record events in your application by sending Errlog the optional
*   tcErr_Mess parameter.
*       Example:   Do ERRLOG with 0, Lineno(), Sys(0), ;
*                            'User pressed Esc while waiting for lock'
*            or:   Do ERRLOG with 0, Lineno(),'','User packed files'
*
*   Calling Errlog to record an event is independent of automatic error
*   logging.  In other words, if you have an "On Error do ERRLOG..." in
*   your program, calling Errlog to log an event will not affect logging of
*   system errors.
*
*   Special Cases:
*   There are three specific errors which Errlog looks for and handles
*   specially.
*      1 - "Structural CDX file not found"  This message can occur from a
*          Use command if the structure CDX has been erased.  Errlog will 
*          issue a RETRY and return.  The data file should be opened with
*          no CDX attached.
*
*      2 - "Not a database file"  This problem occurs when a data file has
*          been corrupted. (Ususally the header's record count doesn't match
*          the actual number of records in the file.)  Errlog will tell the
*          user that a data file has been corrupted and will then quit.
*
*      3 - "Print not available"  Errlog will offer the user three options,
*          Retry the printer, Display only, or Quit the application.  Note that
*          the "Display" option is only available if the error occured on a
*          REPORT FORM command, and if the source code is available.
*
*   All other errors are logged and the message "SYSTEM PROBLEM ENCOUNTERED!
*   CONTACT PROGRAMMER!" is displayed in a wait window.  Errlog will then quit.
*
* Information Recorded:
*   Whenever an error or event is recorded, the following information is
*   saved:
*       Date and time of the error
*       Complete tree of called programs and procedures (in reverse order)
*       Line number in source code (if Lineno() was passed)
*       Source code which caused the error if source code is available and
*           no tcErr_Mess was passed
*       The FoxPro error number, if no tcErr_Mess was passed
*       The FoxPro error message, if no tcErr_Mess was passed
*       ID of user who caused the error, if tcUserID is passed
*       A text listing of record pointers, status, and memory variables
*       Restorable settings of memory variables (excluding system variables)
*       NOTE THAT LOCAL MEMORY VARIABLES IN THE ROUTINE THAT CAUSED THE ERROR
*       CANNOT BE SEEN BY ERRLOG.
*
* Analyzing the log:
*   Errors and events are stored in the file ERRORLOG.DBF.  Use and browse
*   this file to view errors.  Each line in ERRORLOG will show the date and
*   time of the error, the FoxPro error number, and the line of source code
*   which caused the error (if available).  When you find an error you wish
*   to analyze, you can restore the memory from the MEMDUMP field (restore
*   from memo MEMDUMP).
*
*   In addition, the field ERRSTAT shows where the record pointer was for
*   all open files, a status listing, a complete tree of programs and
*   procedures called, and a memory listing.  The ERRSTAT field is in ASCII
*   format and can be viewed with the FoxPro editor or printed.
*
* Reporting log contents:
*   There is a program, screen, and report for reporting a simplified listing
*   of what's in the errorlog.  All three of these components are called
*   Err_Rpt and are as plain vanilla as possible--i.e. they don't rely
*   on any class libraries other than the base VFP classes.  The idea is
*   that you can DO ERR_RPT from the command window if necessary, or call it
*   from a menu option, if you would like to build this reporting capability
*   into your app.
****************************************************************

* Modification history:
* 03/27/02 - Cleaned up variables for VFP standards
* 05/27/94 - Don't exit loop for work areas if empty area found
* 05/04/94 - Still wasn't showing correct selected file in status
* 08/10/93 - Wasn't showing correct selected file in status; If UserID is empty
*            initialize it to 20 spaces instead of null
* 06/18/93 - Only show record pointers for used work areas; Create DBF with
*            SQL command; Show logging in a window; Put called from tree in
*            a memo field; Start all memvars for this routine with ERR_;
*            Change order of parameters and make them all optional
* 02/10/92 - Save error mess to memvar before its destroyed
* 01/29/92 - Increase number of work areas to 25 for Foxpro
* 12/26/90 - Create files in PROBLEMS subdirectory on host for MediMail
* 07/31/90 - DON'T LOG ERROR IF ERROR IS PRINTER NOT READY
* 03/07/90 - CHANGED "SAVE ALL TO..." STATEMENT TO "SAVE ALL LIKE * TO..."
*            FOR COMPATIBILITY WITH FOXPRO
* 03/01/90 - SET PRINT OFF JUST IN CASE PRINTER IS HUNG, WE DON'T WANT
*            SCREEN OUTPUT SENT THERE.  TURN ERROR TRAPPING BACK ON BEFORE
*            RETURNING.  DUKE
* 02/16/90 - ADD APPEND BLANK's BACK IN THAT WERE ACCIDENTALLY DELETED;
*            RESELECT ORIGINAL FILE IF RETURNING
* 02/07/90 - ADDED DISPLAY OF EOF AND BOF TO EACH WORK AREA; ADDED WARNING
*            IF PROBLEMS FILE EXCEEDS 5000 RECORDS
* 02/05/90 - ORIGINAL VERSION

local ldErr_Dt, lcErr_Tm, lnErr_N, lcErr_OFil, lcErr_Mess, loErr_Wind, ;
        lcErr_Call, lnErr_Num, lcErr_Prt, lcErr_Cons, lcErr_Talk, lcCmd, ;
        llErr_Evnt, lcErr_Path, lnErr_DS, lnLineNo, lcUserId, lcAns

* Check for parameters
if empty(tnLineNo)
	lnLineNo = 0
ELSE
	lnLineNo = tnLineNo
endif
if empty(tcUserID)
	lcUserID = ''
ELSE
	lcUserId = tcUserId
endif
if empty(tcErr_Mess)
	llErr_Evnt = .F.           && logging a problem
	lcErr_Mess = message()
else
	llErr_Evnt = .T.           && recording an event
	lcErr_Mess = tcErr_Mess
endif
lnErr_Num = error()

if lnErr_Num = 1707      && STRUCTURAL CDX FILE NOT FOUND
	* This is a workaround for earlier versions of 2.0 where an error
	* would occur on a Use command if the CDX was missing.  Issuing a
	* retry will remove the structure CDX reference from the file and
	* open the database.
	retry
endif

lcErr_Call = on('Error')
on error         && Turn off error logging so it's not called recursively

do case
case empty(tnDataSess) or type('tnDataSess') <> 'N'
	lnErr_DS = 0
case type('tnDataSess') = 'N' and tnDataSess <> 0
	lnErr_DS = set('DataSession')
	set datasession to tnDataSess
case type('thisform.Datasessionid') <> 'U'
	lnErr_Ds = thisform.DatasessionId
otherwise
	lnErr_Ds = 0
endcase

* Save settings in case we are logging an event
lnErr_Prt = set('Print')
set print off
lnErr_Cons = set('Console')
set console on
lnErr_Talk = set('Talk')
set talk off

if type('_screen.activeform') = 'O'
	loErr_Wind = _screen.activeform
else
	loErr_Wind = .null.
endif
ldErr_Dt = date()
lcErr_Tm = time()
lcErr_OFil = select()

if lnErr_Num = 15      && NOT A DATABASE FILE
	clear
	?? chr(7)
	wait 'ONE OF YOUR DATA FILES IS DAMAGED' window
	quit
endif

if lnErr_Num = 125 or lnErr_Num = 1958     && PRINTER NOT READY or Error loading printer driver
	IF NOT EMPTY(MESSAGE(1)) AND "REPORT FORM" $ UPPER(MESSAGE(1))
		DO FORM NoPrinter WITH .T. TO lcAns
	ELSE
		DO FORM NoPrinter TO lcAns
	ENDIF 

	do CASE
	CASE lcAns = "Quit"
		quit
	CASE lcAns = "Retry"
		on error &lcErr_Call
		lcCmd = 'retry'
	OTHERWISE
		* The display option was selected
		lcCmd = MESSAGE(1)+' preview'
		&lcCmd
		&lcCmd = 'return'
	ENDCASE

	on error &lcErr_Call
	set print &lcErr_Prt
	set console &lcErr_Cons
	set talk &lcErr_Talk
	set datasession to lnErr_DS
	&lcCmd
endif

* Turn off the path so we don't find an ERRORLOG.DBF somewhere else.
lcErr_Path = set('Path')
set path to

IF USED('Errorlog')
	SELECT ErrorLog
ELSE 
	select 0
	if not file('ERRORLOG.DBF')
		create table ERRORLOG (DATE D, TIME C (8), ERRNUM N (4), ERRMESS C (79), LINE N (6), SOURCE C (80), OPERATOR C (20), ;
								ERRSTAT M, MEMDUMP M, PROGRAM M)
	ENDIF
	use ERRORLOG
ENDIF 

set path to (lcErr_Path)

activate screen
clear typeahead
if not llErr_Evnt
	?? chr(7)
	Wait 'SYSTEM PROBLEM ENCOUNTERED.  CONTACT PROGRAMMER!...Press a key' window timeout 3
endif
select Errorlog
if reccount()+1 > 20
	wait 'ERROR LOG IS GETTING VERY LARGE.  CONTACT PROGRAMMER IMMEDIATELY!' window timeout 3
endif
append blank
if rlock()
	save to memo MEMDUMP        && save all memory variables
endif
if not llErr_Evnt
	? 'SAVING STATUS AND MEMORY...'
endif
set alternate to ERRTEMP.TXT
set alternate on
if llErr_Evnt
	set console off      && don't want user to see status if logging an event
endif

for lnErr_N = 1 to 2551
	if used(lnErr_N)
		? 'AREA '+ltrim(str(lnErr_N))+': REC# '
		select (lnErr_N)
		?? str(recno(),6)+'  EOF: '+iif(eof(),'.T.','.F.')+'   BOF: '+iif(bof(),'.T.','.F.')
   endif
endfor
select (lcErr_OFil)   && so we can tell which file was selected when we display status
list status
? replicate('*',70)
? 'MEMORY DUMP:'
list memory
? replicate('*',70)
set alternate off
set alternate to
set console on
select ERRORLOG       && record still locked
append memo ERRSTAT from ERRTEMP.TXT sdf
delete file ERRTEMP.TXT
if not llErr_Evnt
	? 'LOGGING PROBLEM...'
endif

replace date with ldErr_Dt, ;
		time with lcErr_Tm, ;
		OPERATOR with lcUserID, ;
		ERRNUM with lnErr_Num, ;
		ERRMESS with lcErr_Mess, ;
		LINE with lnLineNo, ;
		SOURCE with message(1), ;
		ERRSTAT with ERRSTAT+chr(13)+chr(10)+'Calling tree:'+chr(13)+chr(10)+sys(16,1)

lnErr_N = 2
do while len(sys(16,lnErr_N)) <> 0
	* Add each program and procedure called to the ERRSTAT memo field with
	* a carriage return and line feed in between.
	replace ERRSTAT with ERRSTAT+chr(13)+chr(10)+sys(16,lnErr_N)
	lnErr_N = lnErr_N+1
enddo

replace PROGRAM with sys(16,1)
lnErr_N = 2
do while len(sys(16,lnErr_N)) <> 0
	* Add each program and procedure called to the PROGRAM memo field with
	* a carriage return and line feed in between.
	replace PROGRAM with PROGRAM+chr(13)+chr(10)+sys(16,lnErr_N)
	lnErr_N = lnErr_N+1
enddo
use         && close error log file

*  if not an event, exit program
if not llErr_Evnt
	quit
endif

* reselect original file and reset environment
select (lcErr_OFil)
on error &lcErr_Call
set print &lcErr_Prt
set console &lcErr_Cons
set talk &lcErr_Talk
set datasess to lnErr_DS
if not ISNULL(loErr_Wind)
	&loErr_Wind..activate()
endif

return
*-------------------------------------------------------------------------
